unit Data;

//  ////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2006 Jan Punter
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  ////////////////////////////////////////////////////////////////////////////
//
//    lupd : 2006-03-31
//
//  ////////////////////////////////////////////////////////////////////////////


interface

uses

  SysUtils, Classes, StrUtils, Globals, Bithose, Knowledge;


type

  // ///////////////////////////////////////////////////////////////////////////

  TPatchData = class( TObject)
  // Abstract patch data
  private
    FParent    : TPatchData;
    FBinary    : TBitHose;
    FCopyStart : Integer; // A bookmark into the aBitHose parameter for ReadBits members
  private
    function    GetPasName: string;
  private
    procedure   CopyBits ( const aBitHose: TBitHose; aBitOffset, aBitCount: Integer);
    function    ReadBits ( const aBitHose: TBitHose): Boolean; virtual; abstract;
    procedure   WriteBits( const aBitHose: TBitHose);          virtual; abstract;
  public
    constructor Create( aParent: TPatchData);                           virtual;
    destructor  Destroy;                                               override;
    procedure   DumpBinary   ( const aStrings: TStrings; anInd: Integer); virtual;
    procedure   DumpToStrings( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); virtual;
    procedure   DumpDetails  ( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); virtual;
  public
    property    Parent  : TPatchData read FParent;
    property    PasName : string     read GetPasName;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TPatchChunk = class( TPatchData)
  // introduces header type and length
  // Corresponds to a(n abbstract) major section in a patch file.
  private
    FTypeId     : Byte;    // This chunk's type ID
    FDataLength : Word;    // The number of remaining bytes to read for this chunk
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    TypeId     : Byte read FTypeId     write FTypeId;
    property    DataLength : Word read FDataLength write FDataLength;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TPatchDescription = class( TPatchChunk)
  private
    FUnknown1        : TArray7;
    FUnknown2        : TBits5;
    FVoiceCount      : TBits5;
    FBarPosition     : TBits14;
    FUnknown3        : TBits3;
    FRedVisible      : Boolean;
    FBlueVisible     : Boolean;
    FYellowVisible   : Boolean;
    FOrangeVisible   : Boolean;
    FGreenVisible    : Boolean;
    FPurpleVisible   : Boolean;
    FWhiteVisible    : Boolean;
    FMonoPoly        : TBits2;
    FActiveVariation : Byte;
    FCategory        : Byte;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    Unknown1        : TArray7 read FUnknown1        write FUnknown1       ;
    property    Unknown2        : TBits5  read FUnknown2        write FUnknown2       ;
    property    VoiceCount      : TBits5  read FVoiceCount      write FVoiceCount     ;
    property    BarPosition     : TBits14 read FBarPosition     write FBarPosition    ;
    property    Unknown3        : TBits3  read FUnknown3        write FUnknown3       ;
    property    RedVisible      : Boolean read FRedVisible      write FRedVisible     ;
    property    BlueVisible     : Boolean read FBlueVisible     write FBlueVisible    ;
    property    YellowVisible   : Boolean read FYellowVisible   write FYellowVisible  ;
    property    OrangeVisible   : Boolean read FOrangeVisible   write FOrangeVisible  ;
    property    GreenVisible    : Boolean read FGreenVisible    write FGreenVisible   ;
    property    PurpleVisible   : Boolean read FPurpleVisible   write FPurpleVisible  ;
    property    WhiteVisible    : Boolean read FWhiteVisible    write FWhiteVisible   ;
    property    MonoPoly        : TBits2  read FMonoPoly        write FMonoPoly       ;
    property    ActiveVariation : Byte    read FActiveVariation write FActiveVariation;
    property    Category        : Byte    read FCategory        write FCategory       ;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TModule = class( TPatchData)
  private
    FKnowledge   : TModuleKnowledge;
    FModuleIndex : TBits8;
    FRow         : TBits7;
    FCol         : TBits7;
    FColor       : TBits8;
    FUnknown1    : TBits8;
    FUnknown2    : TBits4;
    FModeInfo    : array of TBits6;
  private
    function    GetTypeId          : TBits8;
    function    GetPage            : string;
    function    GetPageIndex       : Integer;
    function    GetShortName       : string;
    function    GetLongName        : string;
    function    GetHeight          : Integer;
    function    GetInputCount      : Integer;
    function    GetOutputCount     : Integer;
    function    GetParamCount      : Integer;
    function    GetModeCount       : Integer;
    function    GetInputTypeCount  : Integer;
    function    GetOutputTypeCount : Integer;
    function    GetParamRangeCount : Integer;
    function    GetModeRangeCount  : Integer;
    function    GetParamTypeCount  : Integer;
    function    GetModeTypeCount   : Integer;
    function    GetInputName ( anIndex: Integer): string;
    function    GetOutputName( anIndex: Integer): string;
    function    GetParamName ( anIndex: Integer): string;
    function    GetModeName  ( anIndex: Integer): string;
  private
    function    GetName: string;
    procedure   SetName( const aValue: string);
    function    GetModeInfoCount: Integer;
    procedure   SetModeInfoCount( aValue: Integer);
    function    GetModeInfo( anIndex: Integer): TBits6;
    procedure   SetModeInfo( anIndex: Integer; aValue: TBits6);
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    destructor  Destroy;                                               override;
    procedure   DumpModes( const aStrings: TStrings; anInd: Integer);
  public
    function    InputsAsString      : string;
    function    OutputsAsString     : string;
    function    ParamsAsString      : string;
    function    ModesAsString       : string;
    function    InputTypesAsString  : string;
    function    OutputTypesAsString : string;
    function    ParamRangesAsString : string;
    function    ModeRangesAsString  : string;
    function    ParamTypesAsString  : string;
    function    ModeTypesAsString   : string;
    function    ConnectorInfo( anIndex: Integer; isInput: Boolean): string;
    procedure   DumpDetails  ( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
    function    ModuleInfo   : string;
  public
    property    Knowledge     : TModuleKnowledge read FKnowledge write FKnowledge;
  public
    property    TypeId          : TBits8  read GetTypeId;
    property    Page            : string  read GetPage;
    property    PageIndex       : Integer read GetPageIndex;
    property    ShortName       : string  read GetShortName;
    property    LongName        : string  read GetLongName;
    property    Height          : Integer read GetHeight;
    property    InputCount      : Integer read GetInputCount;
    property    OutputCount     : Integer read GetOutputCount;
    property    ParamCount      : Integer read GetParamCount;
    property    ModeCount       : Integer read GetModeCount;
    property    InputTypeCount  : Integer read GetInputTypeCount;
    property    OutputTypeCount : Integer read GetOutputTypeCount;
    property    ParamRangeCount : Integer read GetParamRangeCount;
    property    ModeRangeCount  : Integer read GetModeRangeCount;
    property    ParamTypeCount  : Integer read GetParamTypeCount;
    property    ModeTypeCount   : Integer read GetModeTypeCount;
  public
    property    InputName [ anIndex: Integer]: string read GetInputName;
    property    OutputName[ anIndex: Integer]: string read GetOutputName;
    property    ParamName [ anIndex: Integer]: string read GetparamName;
    property    ModeName  [ anIndex: Integer]: string read GetModeName;
  public
    property    ModuleIndex   : TBits8  read FModuleIndex     write FModuleIndex;
    property    Row           : TBits7  read FRow             write FRow        ;
    property    Col           : TBits7  read FCol             write FCol        ;
    property    Color         : TBits8  read FColor           write FColor      ;
    property    Unknown1      : TBits8  read FUnknown1        write FUnknown1   ;
    property    Unknown2      : TBits4  read FUnknown2        write FUnknown2   ;
    property    Name          : string  read GetName          write SetName     ;
    property    ModeInfoCount : Integer read GetModeInfoCount write SetModeInfoCount;
  public
    property    ModeInfo[ anIndex: Integer]: TBits6 read GetModeInfo write SetModeInfo;
  end;

  TModuleClass = class of TModule; // For polymorphic creation, see: CreateModule()


  // ///////////////////////////////////////////////////////////////////////////

  TModuleList = class( TPatchChunk)
  private
    FKnowledge   : TModuleKnowledgeTable;
    FLocation    : TBits2;
    FModuleCount : TBits8;
    FModules     : array of TModule;
  private
    procedure   SetModuleCount( aValue: TBits8);
    function    GetModule( anIndex: Integer): TModule;
    procedure   SetModule( anIndex: Integer; const aValue: TModule);
    procedure   DeleteModules;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    destructor  Destroy;                                               override;
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
    function    FindModule( aModuleIndex: Integer): TModule;
  public
    property    Knowledge: TModuleKnowledgeTable read FKnowledge write FKnowledge;
  public
    property    Location                  : TBits2  read FLocation    write FLocation;
    property    ModuleCount               : TBits8  read FModuleCount write SetModuleCount;
    property    Modules[ anIndex: Integer]: TModule read GetModule    write SetModule;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TUnknown1 = class( TPatchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TCable = class( TPatchData)
  private
    FColor         : TBits3;
    FModuleFrom    : TBits8;
    FConnectorFrom : TBits6;
    FLinkType      : TBits1;
    FModuleTo      : TBits8;
    FConnectorTo   : TBits6;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    function    FindModule( aModuleIndex: TBits8): TModule;
    function    FindModuleInfo( aModuleIndex: TBits8): string;
    function    FindConnectorInfo(
      aModuleIndex   : TBits8;
      aConnectorIndex: TBits6;
      aLinkType      : TBits1;
      isFrom         : Boolean
    ): string;
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    Color         : TBits3 read FColor         write FColor        ;
    property    ModuleFrom    : TBits8 read FModuleFrom    write FModuleFrom   ;
    property    ConnectorFrom : TBits6 read FConnectorFrom write FConnectorFrom;
    property    LinkType      : TBits1 read FLinkType      write FLinkType     ;
    property    ModuleTo      : TBits8 read FModuleTo      write FModuleTo     ;
    property    ConnectorTo   : TBits6 read FConnectorTo   write FConnectorTo  ;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TCableList = class( TPatchChunk)
  private
    FLocation   : TBits2;
    FUnknown    : TBits12;
    FCableCount : TBits10;
    FCables     : array of TCable;
  private
    procedure   SetCableCount( aValue: TBits10);
    function    GetCable( anIndex: Integer): TCable;
    procedure   SetCable( anIndex: Integer; const aValue: TCable);
    procedure   DeleteCables;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    destructor  Destroy;                                               override;
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    Location                 : TBits2  read FLocation   write FLocation;
    property    CableCount               : TBits10 read FCableCount write SetCableCount;
    property    Cables[ anIndex: Integer]: TCable  read GetCable    write SetCable;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TPatchSettings = class( TPatchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TParams = class( TPatchData)
  // Set of parameter values for one module one variation
  private
    FVariation   : Byte;
    FParamValues : array of TBits7;
  private
    function    GetParamCount: Integer;
    procedure   SetParamCount( aValue: Integer);
    function    GetParamValue( anIndex: Integer): TBits7;
    procedure   SetParamValue( anIndex: Integer; aValue: TBits7);
    function    GetParamName( anIndex: Integer): string;
    function    GetParamRange( anIndex: Integer): string;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    Variation  : Byte    read FVariation    write FVariation;
    property    ParamCount : Integer read GetParamCount write SetParamCount;
    property    ParamValue[ anIndex: Integer]: TBits7 read GetParamValue write SetParamValue;
    property    ParamName [ anIndex: Integer]: string read GetParamName;
    property    ParamRange[ anIndex: Integer]: string read GetParamRange;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TParamSet = class( TPatchData)
  // Set of parameter values for one module all variations
  private
    // 0 .. 7 : variations 1 .. 8, 8 : init variation
    FVariations     : array[ 0 .. 8] of TParams;
    FModuleIndex    : TBits8;
    FParameterCount : Integer;
  private
    function    GetVariationCount: Integer;
    function    GetVariation( anIndex: Integer): TParams;
    procedure   SetVariation( anIndex: Integer; aValue: TParams);
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    destructor  Destroy;                                               override;
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
    function    FindModule: TModule;
    function    FindModuleInfo: string;
  public
    property    ModuleIndex    : TBits8  read FModuleIndex    write FModuleIndex;
    property    ParameterCount : Integer read FParameterCount write FParameterCount;
    property    VariationCount : Integer read GetVariationCount;
    property    Variation[ anIndex: Integer]: TParams read GetVariation write SetVariation;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TModuleParameters = class( TPatchChunk)
  private
    FLocation  : Tbits2;
    FUnknown   : TBits8;
    FParamSets : array of TParamSet;
  private
    function    GetSetCount: TBits8;
    procedure   SetSetCount( aValue: TBits8);
    function    GetParamSet( anIndex: TBits8): TParamSet;
    procedure   SetParamSet( anIndex: TBits8; aValue: TParamSet);
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    destructor  Destroy;                                               override;
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    Location : TBits2 read FLocation   write FLocation;
    property    Unknown  : TBits8 read FUnknown    write FUnknown;
    property    SetCount : TBits8 read GetSetCount write SetSetCount;
    property    ParamSet[ anIndex: TBits8]: TParamSet read GetParamSet write SetParamSet;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TUnknown2 = class( TPatchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TKnobAssignments = class( TPatchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TControllerAssignments = class( TPAtchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TUnknown3 = class( TPatchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TName = class( TPatchData)
  private
    FModuleIndex : TBits8;
    FName        : string;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    property    ModuleIndex: TBits8 read FModuleIndex write FModuleIndex;
    property    Name       : string read FName        write FName;
  end;

  // ///////////////////////////////////////////////////////////////////////////

  TModuleNames = class( TPatchChunk)
  private
    FLocation  : TBits2;
    FUnknown   : TBits6;
    FNames     : array of TName;
  private
    function    GetNameCount: Integer;
    procedure   SetNameCount( aValue: Integer);
    function    GetName( aModuleIndex: TBits8): string;
    procedure   SetName( aModuleIndex: TBits8; const aValue: string);
  private
    procedure   AddName( aModuleIndex: TBits8; const aName: string);
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    destructor  Destroy;                                               override;
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    Location  : TBits2 read FLocation    write FLocation;
    property    Unknown   : TBits6 read FUnknown     write FUnknown;
    property    NameCount: Integer read GetNameCount write SetNameCount;
    property    Name[ aModuleIndex: TBits8]: string read GetName write SetName;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TTextPad = class( TPatchChunk)
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TAbstractPatch = class( TPatchData) // patch or performance
  private
    FStrings   : TStringList;
    FBits      : TBitHose;
    FPatchName : string;
    FVersion   : Byte;
    FPatchType : Byte;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
  public
    constructor Create( aParent: TPatchData);                          override;
    destructor  Destroy;                                               override;
  public
    function    LoadFromStream( const aStream   : TStream): Boolean;
    function    SavetoStream  ( const aStream   : TStream): Boolean;
    function    LoadFromFile  ( const aFileName : string ): Boolean;
    function    SavetoFile    ( const aFileName : string ): Boolean;
  public
    procedure   DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    PatchName: string read FpatchName write FPatchName;
    property    Version  : Byte   read FVersion   write FVersion;
    property    PatchType: Byte   read FPatchType write FPatchType;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TPatch = class( TAbstractPatch)
  private
    FModuleKnowledgeTable  : TModuleKnowledgeTable;
    FPatchDescription      : TPatchDescription;
    FModuleListVA          : TModuleList;
    FModuleListFX          : TModuleList;
    FUnknown1              : TUnknown1;
    FCableListVA           : TCableList;
    FCableListFX           : TCableList;
    FPatchSettings         : TPatchSettings;
    FModuleParametersVA    : TModuleParameters;
    FModuleParametersFX    : TModuleParameters;
    FUnknown2              : TUnknown2;
    FKnobAssignments       : TKnobAssignments;
    FControllerAssignments : TControllerAssignments;
    FUnknown3_1            : TUnknown3;
    FUnknown3_2            : TUnknown3;
    FUnknown3_3            : TUnknown3;
    FModuleNamesVA         : TModuleNames;
    FModuleNamesFX         : TModuleNames;
    FTextPad               : TTextPad;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
    procedure   FreeChunks;
  public
    constructor Create( aParent: TPatchData);                          override;
    destructor  Destroy;                                               override;
    function    SearchModule    ( aLocation, aModuleIndex: Integer): TModule;
    function    SearchModuleName( aLocation, aModuleIndex: Integer): string;
    procedure   SetModuleName   ( aLocation, aModuleIndex: Integer; const aValue: string);
  public
    procedure   DumpToStrings( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
    procedure   DumpDetails  ( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    ModuleKnowledgeTable: TModuleKnowledgeTable read FModuleKnowledgeTable write FModuleKnowledgeTable;
  end;


  // ///////////////////////////////////////////////////////////////////////////

  TPerformance = class( TAbstractPatch)
  private
    FPatchA : TPatch;
    FPatchB : TPAtch;
    FPatchC : TPatch;
    FPAtchD : TPatch;
  protected
    function    ReadBits ( const aBitHose: TBitHose): Boolean;         override;
    procedure   WriteBits( const aBitHose: TBitHose);                  override;
    procedure   FreePatches;
  public
    constructor Create( aParent: TPatchData);                          override;
    destructor  Destroy;                                               override;
  public
    procedure   DumpToStrings( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
    procedure   DumpDetails  ( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); override;
  public
    property    PatchA : TPatch read FPatchA;
    property    PatchB : TPatch read FPatchB;
    property    PatchC : TPatch read FPatchC;
    property    PatchD : TPatch read FPatchD;
  end;


// ////////////////////////////////////////////////////////////////////////// //
//
// Module defs.

  TMkeyboard    = class( TModule) end;  TM4out        = class( TModule) end;
  TM2out        = class( TModule) end;  TMInvert      = class( TModule) end;
  TMOscB        = class( TModule) end;  TMOscShpB     = class( TModule) end;
  TMOscC        = class( TModule) end;  TMReverb      = class( TModule) end;
  TMOscString   = class( TModule) end;  TMSw8_1       = class( TModule) end;
  TMValSw1_2    = class( TModule) end;  TMX_Fade      = class( TModule) end;
  TMMix4_1B     = class( TModule) end;  TMEnvADSR     = class( TModule) end;
  TMMux1_8      = class( TModule) end;  TMPartQuant   = class( TModule) end;
  TMModADSR     = class( TModule) end;  TMLfoC        = class( TModule) end;
  TMLfoShpA     = class( TModule) end;  TMLfoA        = class( TModule) end;
  TMOscMaster   = class( TModule) end;  TMSaturate    = class( TModule) end;
  TMMetNoise    = class( TModule) end;  TMDevice      = class( TModule) end;
  TMNoise       = class( TModule) end;  TMEq2Band     = class( TModule) end;
  TMEq3Band     = class( TModule) end;  TMShpExp      = class( TModule) end;
  TMSwOnOffM    = class( TModule) end;  TMPulse       = class( TModule) end;
  TMMix8_1B     = class( TModule) end;  TMEnvH        = class( TModule) end;
  TMDelay       = class( TModule) end;  TMConstant    = class( TModule) end;
  TMLevMult     = class( TModule) end;  TMFltVoice    = class( TModule) end;
  TMEnvAHD      = class( TModule) end;  TMPan         = class( TModule) end;
  TMMixStereo   = class( TModule) end;  TMFltMulti    = class( TModule) end;
  TMConstSwT    = class( TModule) end;  TMFltNord     = class( TModule) end;
  TMEnvMulti    = class( TModule) end;  TMS_and_H     = class( TModule) end;
  TMFltStatic   = class( TModule) end;  TMEnvD        = class( TModule) end;
  TMAutomate    = class( TModule) end;  TMDrumSynth   = class( TModule) end;
  TMCompLev     = class( TModule) end;  TMMux8_1X     = class( TModule) end;
  TMClip        = class( TModule) end;  TMOverdrive   = class( TModule) end;
  TMScratch     = class( TModule) end;  TMGate        = class( TModule) end;
  TMMix2_1B     = class( TModule) end;  TMClkGen      = class( TModule) end;
  TMClkDiv      = class( TModule) end;  TMEnvFollow   = class( TModule) end;
  TMNoteScaler  = class( TModule) end;  TMWaveWrap    = class( TModule) end;
  TMNoteQuant   = class( TModule) end;  TMSwOnOffT    = class( TModule) end;
  TMSw1_8       = class( TModule) end;  TMSw4_1       = class( TModule) end;
  TMLevAmp      = class( TModule) end;  TMRect        = class( TModule) end;
  TMShpStatic   = class( TModule) end;  TMEnvADR      = class( TModule) end;
  TMWindSw      = class( TModule) end;  TM8Counter    = class( TModule) end;
  TMFltLP       = class( TModule) end;  TMSw1_4       = class( TModule) end;
  TMFlanger     = class( TModule) end;  TMSw1_2       = class( TModule) end;
  TMFlipFlop    = class( TModule) end;  TMFltClassic  = class( TModule) end;
  TMStChorus    = class( TModule) end;  TMOscD        = class( TModule) end;
  TMOscA        = class( TModule) end;  TMFreqShift   = class( TModule) end;
  TMSw2_1       = class( TModule) end;  TMFltPhase    = class( TModule) end;
  TMEqPeak      = class( TModule) end;  TMValSw2_1    = class( TModule) end;
  TMOscNoise    = class( TModule) end;  TMVocoder     = class( TModule) end;
  TMLevAdd      = class( TModule) end;  TMFade1_2     = class( TModule) end;
  TMFade2_1     = class( TModule) end;  TMLevScaler   = class( TModule) end;
  TMMix8_1A     = class( TModule) end;  TMLevMod      = class( TModule) end;
  TMDigitizer   = class( TModule) end;  TMEnvADDSR    = class( TModule) end;
  TMSeqNote     = class( TModule) end;  TMMix4_1C     = class( TModule) end;
  TMMux8_1      = class( TModule) end;  TMWahWah      = class( TModule) end;
  TMName        = class( TModule) end;  TMFx_In       = class( TModule) end;
  TMMinMax      = class( TModule) end;  TMBinCounter  = class( TModule) end;
  TMADConv      = class( TModule) end;  TMDAConv      = class( TModule) end;
  TMFltHP       = class( TModule) end;  TMT_and_H     = class( TModule) end;
  TMMix4_1S     = class( TModule) end;  TMCtrlSend    = class( TModule) end;
  TMPCSend      = class( TModule) end;  TMNoteSend    = class( TModule) end;
  TMSeqEvent    = class( TModule) end;  TMSeqValue    = class( TModule) end;
  TMSeqLev      = class( TModule) end;  TMCtrlRcv     = class( TModule) end;
  TMNoteRcv     = class( TModule) end;  TMNoteZone    = class( TModule) end;
  TMCompressor  = class( TModule) end;  TMKeyQuant    = class( TModule) end;
  TMSeqCtr      = class( TModule) end;  TMNoteDet     = class( TModule) end;
  TMLevConv     = class( TModule) end;  TMGlide       = class( TModule) end;
  TMCompSig     = class( TModule) end;  TMZeroCnt     = class( TModule) end;
  TMMixFader    = class( TModule) end;  TMFltComb     = class( TModule) end;
  TMOscShpA     = class( TModule) end;  TMOscDual     = class( TModule) end;
  TMDXRouter    = class( TModule) end;  TMPShift      = class( TModule) end;
  TMModAHD      = class( TModule) end;  TM2_In        = class( TModule) end;
  TM4_In        = class( TModule) end;  TMDlySingleA  = class( TModule) end;
  TMDlySingleB  = class( TModule) end;  TMDelayDual   = class( TModule) end;
  TMDelayQuad   = class( TModule) end;  TMDelayA      = class( TModule) end;
  TMDelayB      = class( TModule) end;  TMDlyClock    = class( TModule) end;
  TMDlyShiftReg = class( TModule) end;  TMOperator    = class( TModule) end;
  TMDlyEight    = class( TModule) end;  TMDlyStereo   = class( TModule) end;
  TMOscPM       = class( TModule) end;  TMMix1_1A     = class( TModule) end;
  TMMix1_1S     = class( TModule) end;  TMSw1_2M      = class( TModule) end;
  TMSw2_1M      = class( TModule) end;  TMConstSwM    = class( TModule) end;
  TMNoiseGate   = class( TModule) end;  TMLfoB        = class( TModule) end;
  TMPhaser      = class( TModule) end;  TMMix4_1A     = class( TModule) end;
  TMMix2_1A     = class( TModule) end;  TMModAmt      = class( TModule) end;
  TMOscPerc     = class( TModule) end;  TMStatus      = class( TModule) end;
  TMPitchTrack  = class( TModule) end;  TMMonoKey     = class( TModule) end;
  TMRandomA     = class( TModule) end;  TMRandomB     = class( TModule) end;
  TMRndClkA     = class( TModule) end;  TMRndTrig     = class( TModule) end;
  TMRndClkB     = class( TModule) end;  TMRndPattern  = class( TModule) end;


// End - module defs
//
// ////////////////////////////////////////////////////////////////////////// //



implementation

uses Math;


type

  TMTMap = record        // T_Module_Type_Map
    mt : Integer;        // Module type ID
    mc : TModuleClass; // Module class
  end;


const

  TMTMaps : array[ 0 .. 165] of TMTMap =
  (
    ( mt:   1; mc : TMKeyboard   ), ( mt:   3; mc : TM4out       ), ( mt:   4; mc : TM2out       ),
    ( mt:   5; mc : TMInvert     ), ( mt:   7; mc : TMOscB       ), ( mt:   8; mc : TMOscShpB    ),
    ( mt:   9; mc : TMOscC       ), ( mt:  12; mc : TMReverb     ), ( mt:  13; mc : TMOscString  ),
    ( mt:  15; mc : TMSw8_1      ), ( mt:  17; mc : TMValSw1_2   ), ( mt:  18; mc : TMX_Fade     ),
    ( mt:  19; mc : TMMix4_1B    ), ( mt:  20; mc : TMEnvADSR    ), ( mt:  21; mc : TMMux1_8     ),
    ( mt:  22; mc : TMPartQuant  ), ( mt:  23; mc : TMModADSR    ), ( mt:  24; mc : TMLfoC       ),
    ( mt:  25; mc : TMLfoShpA    ), ( mt:  26; mc : TMLfoA       ), ( mt:  27; mc : TMOscMaster  ),
    ( mt:  28; mc : TMSaturate   ), ( mt:  29; mc : TMMetNoise   ), ( mt:  30; mc : TMDevice     ),
    ( mt:  31; mc : TMNoise      ), ( mt:  32; mc : TMEq2Band    ), ( mt:  33; mc : TMEq3Band    ),
    ( mt:  34; mc : TMShpExp     ), ( mt:  36; mc : TMSwOnOffM   ), ( mt:  38; mc : TMPulse      ),
    ( mt:  40; mc : TMMix8_1B    ), ( mt:  41; mc : TMEnvH       ), ( mt:  42; mc : TMDelay      ),
    ( mt:  43; mc : TMConstant   ), ( mt:  44; mc : TMLevMult    ), ( mt:  45; mc : TMFltVoice   ),
    ( mt:  46; mc : TMEnvAHD     ), ( mt:  47; mc : TMPan        ), ( mt:  48; mc : TMMixStereo  ),
    ( mt:  49; mc : TMFltMulti   ), ( mt:  50; mc : TMConstSwT   ), ( mt:  51; mc : TMFltNord    ),
    ( mt:  52; mc : TMEnvMulti   ), ( mt:  53; mc : TMS_and_H    ), ( mt:  54; mc : TMFltStatic  ),
    ( mt:  55; mc : TMEnvD       ), ( mt:  57; mc : TMAutomate   ), ( mt:  58; mc : TMDrumSynth  ),
    ( mt:  59; mc : TMCompLev    ), ( mt:  60; mc : TMMux8_1X    ), ( mt:  61; mc : TMClip       ),
    ( mt:  62; mc : TMOverdrive  ), ( mt:  63; mc : TMScratch    ), ( mt:  64; mc : TMGate       ),
    ( mt:  66; mc : TMMix2_1B    ), ( mt:  68; mc : TMClkGen     ), ( mt:  69; mc : TMClkDiv     ),
    ( mt:  71; mc : TMEnvFollow  ), ( mt:  72; mc : TMNoteScaler ), ( mt:  74; mc : TMWaveWrap   ),
    ( mt:  75; mc : TMNoteQuant  ), ( mt:  76; mc : TMSwOnOffT   ), ( mt:  78; mc : TMSw1_8      ),
    ( mt:  79; mc : TMSw4_1      ), ( mt:  81; mc : TMLevAmp     ), ( mt:  82; mc : TMRect       ),
    ( mt:  83; mc : TMShpStatic  ), ( mt:  84; mc : TMEnvADR     ), ( mt:  85; mc : TMWindSw     ),
    ( mt:  86; mc : TM8Counter   ), ( mt:  87; mc : TMFltLP      ), ( mt:  88; mc : TMSw1_4      ),
    ( mt:  89; mc : TMFlanger    ), ( mt:  90; mc : TMSw1_2      ), ( mt:  91; mc : TMFlipFlop   ),
    ( mt:  92; mc : TMFltClassic ), ( mt:  94; mc : TMStChorus   ), ( mt:  96; mc : TMOscD       ),
    ( mt:  97; mc : TMOscA       ), ( mt:  98; mc : TMFreqShift  ), ( mt: 100; mc : TMSw2_1      ),
    ( mt: 102; mc : TMFltPhase   ), ( mt: 103; mc : TMEqPeak     ), ( mt: 105; mc : TMValSw2_1   ),
    ( mt: 106; mc : TMOscNoise   ), ( mt: 108; mc : TMVocoder    ), ( mt: 112; mc : TMLevAdd     ),
    ( mt: 113; mc : TMFade1_2    ), ( mt: 114; mc : TMFade2_1    ), ( mt: 115; mc : TMLevScaler  ),
    ( mt: 116; mc : TMMix8_1A    ), ( mt: 117; mc : TMLevMod     ), ( mt: 118; mc : TMDigitizer  ),
    ( mt: 119; mc : TMEnvADDSR   ), ( mt: 121; mc : TMSeqNote    ), ( mt: 123; mc : TMMix4_1C    ),
    ( mt: 124; mc : TMMux8_1     ), ( mt: 125; mc : TMWahWah     ), ( mt: 126; mc : TMName       ),
    ( mt: 127; mc : TMFx_In      ), ( mt: 128; mc : TMMinMax     ), ( mt: 130; mc : TMBinCounter ),
    ( mt: 131; mc : TMADConv     ), ( mt: 132; mc : TMDAConv     ), ( mt: 134; mc : TMFltHP      ),
    ( mt: 139; mc : TMT_and_H    ), ( mt: 140; mc : TMMix4_1S    ), ( mt: 141; mc : TMCtrlSend   ),
    ( mt: 142; mc : TMPCSend     ), ( mt: 143; mc : TMNoteSend   ), ( mt: 144; mc : TMSeqEvent   ),
    ( mt: 145; mc : TMSeqValue   ), ( mt: 146; mc : TMSeqLev     ), ( mt: 147; mc : TMCtrlRcv    ),
    ( mt: 148; mc : TMNoteRcv    ), ( mt: 149; mc : TMNoteZone   ), ( mt: 150; mc : TMCompressor ),
    ( mt: 152; mc : TMKeyQuant   ), ( mt: 154; mc : TMSeqCtr     ), ( mt: 156; mc : TMNoteDet    ),
    ( mt: 157; mc : TMLevConv    ), ( mt: 158; mc : TMGlide      ), ( mt: 159; mc : TMCompSig    ),
    ( mt: 160; mc : TMZeroCnt    ), ( mt: 161; mc : TMMixFader   ), ( mt: 162; mc : TMFltComb    ),
    ( mt: 163; mc : TMOscShpA    ), ( mt: 164; mc : TMOscDual    ), ( mt: 165; mc : TMDXRouter   ),
    ( mt: 167; mc : TMPShift     ), ( mt: 169; mc : TMModAHD     ), ( mt: 170; mc : TM2_In       ),
    ( mt: 171; mc : TM4_In       ), ( mt: 172; mc : TMDlySingleA ), ( mt: 173; mc : TMDlySingleB ),
    ( mt: 174; mc : TMDelayDual  ), ( mt: 175; mc : TMDelayQuad  ), ( mt: 176; mc : TMDelayA     ),
    ( mt: 177; mc : TMDelayB     ), ( mt: 178; mc : TMDlyClock   ), ( mt: 179; mc : TMDlyShiftReg),
    ( mt: 180; mc : TMOperator   ), ( mt: 181; mc : TMDlyEight   ), ( mt: 182; mc : TMDlyStereo  ),
    ( mt: 183; mc : TMOscPM      ), ( mt: 184; mc : TMMix1_1A    ), ( mt: 185; mc : TMMix1_1S    ),
    ( mt: 186; mc : TMSw1_2M     ), ( mt: 187; mc : TMSw2_1M     ), ( mt: 188; mc : TMConstSwM   ),
    ( mt: 189; mc : TMNoiseGate  ), ( mt: 190; mc : TMLfoB       ), ( mt: 192; mc : TMPhaser     ),
    ( mt: 193; mc : TMMix4_1A    ), ( mt: 194; mc : TMMix2_1A    ), ( mt: 195; mc : TMModAmt     ),
    ( mt: 196; mc : TMOscPerc    ), ( mt: 197; mc : TMStatus     ), ( mt: 198; mc : TMPitchTrack ),
    ( mt: 199; mc : TMMonoKey    ), ( mt: 200; mc : TMRandomA    ), ( mt: 202; mc : TMRandomB    ),
    ( mt: 204; mc : TMRndClkA    ), ( mt: 205; mc : TMRndTrig    ), ( mt: 206; mc : TMRndClkB    ),
    ( mt: 208; mc : TMRndPattern )
  );


  function CreateModule( aParent: TPatchData; aType: Byte): TModule;
  var
    i : Integer;
  begin
    Result := nil;
    for i := Low( TMTMaps) to High( TMTMaps)
    do
      if TMTMaps[ i].mt = aType
      then begin
        Result := TMTMaps[ i].mc.Create( aParent);
        break;
      end;
    if not Assigned( Result)
    then Result := TModule.Create( aParent);
  end;




{ ========
  TPatchData = class( TObject)
  private
    FParent     : TPatchData;
    FBinary     : TBitHose;
    FCopyStart  : Integer;
  public
    property    Parent : TPatchData read FParent;
  protected

    // function    ReadBits ( const aBitHose: TBitHose): Boolean; // virtual; abstract;
    // procedure   WriteBits( const aBitHose: TBitHose);          // virtual; abstract;

}

    procedure   TPatchData.CopyBits( const aBitHose: TBitHose; aBitOffset, aBitCount: Integer);
    begin
      FBinary.CopyFrom( aBitHose, aBitOffset, aBitCount);
    end;


//  private

    function    TPatchData.GetPasName: string;
    begin
      Result := Copy( ClassName, 2, Length( ClassName) - 1);
    end;

//  public

    constructor TPatchData.Create( aParent: TPatchData);
    begin
      inherited Create;
      FParent  := aParent;
      FBinary  := TBitHose.Create;
    end;

    destructor  TPatchData.Destroy; // override;
    begin
      FreeAndNil( FBinary);
      inherited;
    end;

//  public

    procedure   TPatchData.DumpBinary( const aStrings: TStrings; anInd: Integer); // virtual;
    const
      BlockSize = 8; // Best a multiple of two, I guess
    var
      i : Integer;
      S : string;
      F : Boolean;
    begin
      S := Format( '%s// %.4x : ', [ Indent( anInd + 1), 0]);
      F := False;
      for i := 0 to ( FBinary.Size + 7) div 8 - 1
      do begin
        F := true;
        S := S + Format( ' #%s', [ ByteToBinary( FBinary.Bits[ 8 * i, 8], 8)]);
        if (( i + BlockSize div 2 + 1) mod BlockSize) = 0
        then S := S + ' ';
        if (( i + 1) mod BlockSize) = 0
        then begin
          aStrings.Add( S);
          S := Format( '%s// %.4x : ', [ Indent( anInd + 1), ( i + 1) div BlockSize * BlockSize]);
          F := False;
        end;
      end;
      if F
      then aStrings.Add( S);
    end;

    procedure   TPatchData.DumpToStrings( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // virtual;
    begin
      with aStrings
      do begin
        Add( IndFmt( anInd, '%s', [ PasName]));
        Add( IndFmt( anInd, '[' , [ ]));
        if aShowBinary
        then DumpBinary( aStrings, anInd);
        DumpDetails( aStrings, anInd + 1, aShowBinary, aShowKnowledge);
        Add( IndFmt( anInd, '] // %s', [ PasName]));
      end;
    end;

    procedure   TPatchData.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // virtual;
    begin
      with aStrings
      do begin
        Add( IndFmt( anInd, '// PatchData', []));
        Add( IndFmt( anInd, 'class = %s'  , [ PasName]))
      end;
    end;



{ ========
  TPatchChunk = class( TPatchData)
  // introduces header type and length
  private
    FTypeId     : Byte;
    FDataLength : Word;
  public
    property    TypeId     : Byte read FTypeId     write FTypeId;
    property    DataLength : Word read FDataLength write FDataLength;
  protected
}

    function    TPatchChunk.ReadBits( const aBitHose: TBitHose): Boolean; // override;
    begin
      Result := False;
      try
        with aBitHose
        do begin
          FCopyStart := ReadPointer;
          Result :=
            ( TypeId     = Read(  8)) and
            ( DataLength = Read( 16));
        end;
        CopyBits( aBitHose, FCopyStart, 24 + DataLength * 8);
        FCopyStart := FCopyStart+ 24;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TPatchChunk.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;


//  public

    procedure   TPatchChunk.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      if aShowBinary
      then
        with aStrings
        do begin
          Add( IndFmt( anInd, 'TypeId     = $%.2x', [ TypeId     ]));
          Add( IndFmt( anInd, 'DataLength = %d'   , [ DataLength ]));
          Add( IndFmt( anInd, '// ---- data -----', [            ]));
        end;
    end;



{ ========
  TPatchDescription = class( TPatchChunk)
  private
    FUnknown1        : TArray7;
    FUnknown2        : TBits5;
    FVoiceCount      : TBits5;
    FBarPosition     : TBits14;
    FUnknown3        : TBits3;
    FRedVisible      : Boolean;
    FBlueVisible     : Boolean;
    FYellowVisible   : Boolean;
    FOrangeVisible   : Boolean;
    FGreenVisible    : Boolean;
    FPurpleVisible   : Boolean;
    FWhiteVisible    : Boolean;
    FMonoPoly        : TBits2;
    FActiveVariation : Byte;
    FCategory        : Byte;
  public
    property    Unknown1        : TArray7 read FUnknown1        write FUnknown1       ;
    property    Unknown2        : TBits5  read FUnknown2        write FUnknown2       ;
    property    VoiceCount      : TBits5  read FVoiceCount      write FVoiceCount     ;
    property    BarPosition     : TBits14 read FBarPosition     write FBarPosition    ;
    property    Unknown3        : TBits3  read FUnknown3        write FUnknown3       ;
    property    RedVisible      : Boolean read FRedVisible      write FRedVisible     ;
    property    BlueVisible     : Boolean read FBlueVisible     write FBlueVisible    ;
    property    YellowVisible   : Boolean read FYellowVisible   write FYellowVisible  ;
    property    OrangeVisible   : Boolean read FOrangeVisible   write FOrangeVisible  ;
    property    GreenVisible    : Boolean read FGreenVisible    write FGreenVisible   ;
    property    PurpleVisible   : Boolean read FPurpleVisible   write FPurpleVisible  ;
    property    WhiteVisible    : Boolean read FWhiteVisible    write FWhiteVisible   ;
    property    MonoPoly        : TBots2  read FMonoPoly        write FMonoPoly       ;
    property    ActiveVariation : Byte    read FActiveVariation write FActiveVariation;
    property    Category        : Byte    read FCategory        write FCategory       ;
  protected
}


    function    TPatchDescription.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      i    : Integer;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
          for i := 0 to 6
          do FUnknown1[ i] := Read(  8);
          FUnknown2        := Read(  5);
          VoiceCount       := Read(  5);        
          BarPosition      := Read( 14);        
          Unknown3         := Read(  3);        
          RedVisible       := Read(  1) <> 0;   
          BlueVisible      := Read(  1) <> 0;   
          YellowVisible    := Read(  1) <> 0;   
          OrangeVisible    := Read(  1) <> 0;   
          GreenVisible     := Read(  1) <> 0;   
          PurpleVisible    := Read(  1) <> 0;   
          WhiteVisible     := Read(  1) <> 0;   
          MonoPoly         := Read(  2);        
          ActiveVariation  := Read(  8);        
          Category         := Read(  8);
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TPatchDescription.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TPatchDescription.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited DumpDetails( aStrings, anInd, aShowBinary, aShowKnowledge);
      with aStrings
      do begin
        Add( IndFmt( anInd, 'Unknown1        = [%s]', [ BytesToHex      ( Unknown1      ) ]));
        Add( IndFmt( anInd, 'Unknown2        = #%s' , [ ByteToBinary    ( Unknown2, 5   ) ]));
        Add( IndFmt( anInd, 'VoiceCount      = %d'  , [ VoiceCount                        ]));
        Add( IndFmt( anInd, 'BarPosition     = %d'  , [ BarPosition                       ]));
        Add( IndFmt( anInd, 'Unknown3        = #%s' , [ ByteToBinary    ( Unknown2, 3   ) ]));
        Add( IndFmt( anInd, 'RedVisible      = %s'  , [ Bools           ( RedVisible    ) ]));
        Add( IndFmt( anInd, 'BlueVisible     = %s'  , [ Bools           ( BlueVisible   ) ]));
        Add( IndFmt( anInd, 'YellowVisible   = %s'  , [ Bools           ( YellowVisible ) ]));
        Add( IndFmt( anInd, 'OrangeVisible   = %s'  , [ Bools           ( OrangeVisible ) ]));
        Add( IndFmt( anInd, 'GreenVisible    = %s'  , [ Bools           ( GreenVisible  ) ]));
        Add( IndFmt( anInd, 'PurpleVisible   = %s'  , [ Bools           ( PurpleVisible ) ]));
        Add( IndFmt( anInd, 'WhiteVisible    = %s'  , [ Bools           ( WhiteVisible  ) ]));
        Add( IndFmt( anInd, 'MonoPoly        = %s'  , [ MonoPolies      ( MonoPoly      ) ]));
        Add( IndFmt( anInd, 'ActiveVariation = %d'  , [ ActiveVariation                   ]));
        Add( IndFmt( anInd, 'Category        = %s'  , [ CatName         ( Category      ) ]));
      end;
    end;


{ ========
  TModule = class( TPatchData)
  private
    FKnowledge   : TModuleKnowledge;
    FModuleIndex : TBits8;
    FRow         : TBits7;
    FCol         : TBits7;
    FColor       : TBits8;
    FUnknown1    : TBits8;
    FUnknown2    : TBits4;
    FModeInfo    : array of TBits6;
  public
    property    Knowledge     : TModuleKnowledge read FKnowledge write FKnowledge;
  public
    property    TypeId          : TBits8  read GetTypeId;
    property    Page            : string  read GetPage;
    property    PageIndex       : Integer read GetPageIndex;
    property    ShortName       : string  read GetShortName;
    property    LongName        : string  read GetLongName;
    property    Height          : Integer read GetHeight;
    property    InputCount      : Integer read GetInputCount;
    property    OutputCount     : Integer read GetOutputCount;
    property    ParamCount      : Integer read GetParamCount;
    property    ModeCount       : Integer read GetModeCount;
    property    InputTypeCount  : Integer read GetInputTypeCount;
    property    OutputTypeCount : Integer read GetOutputTypeCount;
    property    ParamRangeCount : Integer read GetParamRangeCount;
    property    ModeRangeCount  : Integer read GetModeRangeCount;
  public
    property    InputName [ anIndex: Integer]: string read GetInputName;
    property    OutputName[ anIndex: Integer]: string read GetOutputName;
    property    ParamName [ anIndex: Integer]: string read GetparamName;
    property    ModeName  [ anIndex: Integer]: string read GetModeName;
  public
    property    ModuleIndex   : TBits8  read FModuleIndex     write FModuleIndex;
    property    Row           : TBits7  read FRow             write FRow        ;
    property    Col           : TBits7  read FCol             write FCol        ;
    property    Color         : TBits8  read FColor           write FColor      ;
    property    Unknown1      : TBits8  read FUnknown1        write FUnknown1   ;
    property    Unknown2      : TBits4  read FUnknown2        write FUnknown2   ;
    property    Name          : string  read GetName          write SetName     ;
    property    ModeInfoCount : Integer read GetModeInfoCount write SetModeInfoCount;
  public
    property    ModeInfo[ anIndex: Integer]: TBits6 read GetModeInfo write SetModeInfo;
  private
}

    function    TModule.GetTypeId : TBits8;
    begin
      Result := Knowledge.TypeId;
    end;

    function    TModule.GetPage: string;
    begin
      Result := Knowledge.Page;
    end;

    function    TModule.GetPageIndex: Integer;
    begin
      Result := Knowledge.PageIndex;
    end;

    function    TModule.GetShortName: string;
    begin
      Result := Knowledge.ShortName;
    end;

    function    TModule.GetLongName: string;
    begin
      result := Knowledge.LongName;
    end;

    function    TModule.GetHeight: Integer;
    begin
      Result := Knowledge.Height;
    end;

    function    TModule.GetInputCount: Integer;
    begin
      Result := Knowledge.InputCount;
    end;

    function    TModule.GetOutputCount: Integer;
    begin
      Result := Knowledge.OutputCount;
    end;

    function    TModule.GetParamCount: Integer;
    begin
      Result := Knowledge.ParamCount;
    end;

    function    TModule.GetModeCount: Integer;
    begin
      Result := Knowledge.ModeCount;
    end;

    function    TModule.GetInputTypeCount: Integer;
    begin
      Result := Knowledge.InputTypeCount;
    end;

    function    TModule.GetOutputTypeCount: Integer;
    begin
      Result := Knowledge.OutputTypeCount;
    end;

    function    TModule.GetParamRangeCount: Integer;
    begin
      Result := Knowledge.ParamRangeCount;
    end;

    function    TModule.GetModeRangeCount: Integer;
    begin
      Result := Knowledge.ModeRangeCount;
    end;

    function    TModule.GetParamTypeCount: Integer;
    begin
      Result := Knowledge.ParamTypeCount;
    end;

    function    TModule.GetModeTypeCount: Integer;
    begin
      Result := Knowledge.ModeTypeCount;
    end;

    function    TModule.GetInputName( anIndex: Integer): string;
    begin
      Result := Knowledge.InputName[ anIndex];
    end;

    function    TModule.GetOutputName( anIndex: Integer): string;
    begin
      Result := Knowledge.OutputName[ anIndex]
    end;

    function    TModule.GetParamName( anIndex: Integer): string;
    begin
      Result := Knowledge.ParamName[ anIndex]
    end;

    function    TModule.GetModeName( anIndex: Integer): string;
    begin
      Result := Knowledge.ModeName[ anIndex]
    end;

  // private

    function    TModule.GetName: string;
    var
      aLocation : Integer;
      aPatch    : TPatch;
    begin
      Result := 'module_name_not_found';
      if Assigned( Parent) and ( Parent is TModuleList)
      then begin
        aLocation := TModuleList( Parent).Location;
        aPatch    := Parent.Parent as TPatch;
        if Assigned( aPatch)
        then Result := aPatch.SearchModuleName( aLocation, ModuleIndex);
      end;
    end;

    procedure   TModule.SetName( const aValue: string);
    var
      aLocation : Integer;
      aPatch    : TPatch;
    begin
      if Assigned( FParent) and ( FParent is TModuleList)
      then begin
        aLocation := TModuleList( FParent).Location;
        aPatch    := FParent.Parent as TPatch;
        if Assigned( aPatch)
        then aPatch.SetModuleName( aLocation, ModuleIndex, aValue);
      end;
    end;

    function    TModule.GetModeInfoCount: Integer;
    begin
      Result := Length( FModeInfo);
    end;

    procedure   TModule.SetModeInfoCount( aValue: Integer);
    begin
      SetLength( FModeInfo, aValue);
    end;

    function    TModule.GetModeInfo( anIndex: Integer): TBits6;
    begin
      if ( anIndex >= 0) and ( anIndex < ModeInfoCount)
      then Result := FModeInfo[ anIndex]
      else Result := High( TBits6);
    end;

    procedure   TModule.SetModeInfo( anIndex: Integer; aValue: TBits6);
    begin
      if ( anIndex >= 0) and ( anIndex < ModeInfoCount)
      then FModeInfo[ anIndex] := aValue;
    end;


//  protected

    function    TModule.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      i : Integer;
    begin
      Result := False;
      try
        with aBitHose
        do begin
          if TypeId = Read( 8)
          then begin
            ModuleIndex   := Read( 8);
            Col           := Read( 7);
            Row           := Read( 7);
            Color         := Read( 8);
            Unknown1      := Read( 8);
            Unknown2      := Read( 4);
            ModeInfoCount := ModeCount;
            for i := 0 to ModeCount - 1
            do ModeInfo[ i] := Read( 6);
            Result := True;
          end
          else Result := False;
        end;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TModule.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    destructor  TModule.Destroy; // override;
    begin
      ModeInfoCount := 0;
      inherited;
    end;

    procedure   TModule.DumpModes( const aStrings: TStrings; anInd: Integer);
    var
      i : Integer;
    begin
      with aStrings
      do begin
        Add( IndStr( anInd,   'Modes['));
        for i := 0 to ModeInfoCount - 1
        do begin
          Add( IndFmt( anInd, '  ModeVal %d = %d // %s :: %s', [ i, ModeInfo[ i], ExtendStr( ModeName[ i], 12), Knowledge.ModeRange[ i] ]));
        end;
        Add( IndStr( anInd,   '] // Modes'));
      end;
    end;

    function    TModule.InputsAsString: string;
    begin
      Result := Knowledge.InputsAsString;
    end;

    function    TModule.OutputsAsString: string;
    begin
      Result := Knowledge.OutputsAsString;
    end;

    function    TModule.ParamsAsString: string;
    begin
      Result := Knowledge.ParamsAsString;
    end;

    function    TModule.ModesAsString: string;
    begin
      Result := Knowledge.ModesAsString;
    end;

    function    TModule.InputTypesAsString: string;
    begin
      Result := Knowledge.InputTypesAsString;
    end;

    function    TModule.OutputTypesAsString: string;
    begin
      Result := Knowledge.OutputTypesAsString;
    end;

    function    TModule.ParamRangesAsString: string;
    begin
      Result := Knowledge.ParamRangesAsString;
    end;

    function    TModule.ModeRangesAsString: string;
    begin
      Result := Knowledge.ModeRangesAsString;
    end;

    function    TModule.ParamTypesAsString: string;
    begin
      Result := Knowledge.ParamTypesAsString;
    end;

    function    TModule.ModeTypesAsString: string;
    begin
      Result := Knowledge.ModeTypesAsString;
    end;

    function    TModule.ConnectorInfo( anIndex: Integer; isInput: Boolean): string; // override;
    begin
      Result := Knowledge.ConnectorInfo( anIndex, isInput);
    end;


    procedure   TModule.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
      function MakeMsg( aCnt: Integer): string; begin Result := Format( '*** COUNT ERROR (%d) * ', [ aCnt]); end;
      function Select( aCnt1, aCnt2, aCnt3: Integer): string; begin Result := IfThen(( aCnt1 = aCnt2) and ( aCnt1 = aCnt3), '', MakeMsg( aCnt1))end;
      function CheckInputs  : string; begin Result := Select( InputCount  , InputTypeCount , InputCount    ); end;
      function CheckOutputs : string; begin Result := Select( OutputCount , OutputTypeCount, OutputCount   ); end;
      function CheckParams  : string; begin Result := Select( ParamCount  , ParamRangeCount, ParamTypeCount); end;
      function CheckModes   : string; begin Result := Select( ModeCount   , ModeRangeCount , ModeTypeCount ); end;
    begin
      with aStrings
      do begin
        if aShowKnowledge
        then begin
          Add( IndFmt( anInd,   'Knowledge['            , [                                   ] ));
          Add( IndFmt( anInd,   '  Page        = ''%s''', [ Page                              ] ));
          Add( IndFmt( anInd,   '  PageIndex   = %d'    , [ PageIndex                         ] ));
          Add( IndFmt( anInd,   '  ShortName   = ''%s''', [ ShortName                         ] ));
          Add( IndFmt( anInd,   '  LongName    = ''%s''', [ LongName                          ] ));
          Add( IndFmt( anInd,   '  Height      = %d'    , [ Height                            ] ));
          Add( IndFmt( anInd, '%s  Inputs      = %s'    , [ CheckInputs , InputsAsString      ] ));
          Add( IndFmt( anInd, '%s  InputTypes  = %s'    , [ CheckInputs , InputTypesAsString  ] ));
          Add( IndFmt( anInd, '%s  Outputs     = %s'    , [ CheckOutputs, OutputsAsString     ] ));
          Add( IndFmt( anInd, '%s  OutputTypes = %s'    , [ CheckOutputs, OutputTypesAsString ] ));
          Add( IndFmt( anInd, '%s  Params      = %s'    , [ CheckParams , ParamsAsString      ] ));
          Add( IndFmt( anInd, '%s  ParamRanges = %s'    , [ CheckParams , ParamRangesAsString ] ));
          Add( IndFmt( anInd, '%s  ParamTypes  = %s'    , [ CheckParams , ParamTypesAsString  ] ));
          Add( IndFmt( anInd, '%s  Modes       = %s'    , [ CheckModes  , ModesAsString       ] ));
          Add( IndFmt( anInd, '%s  ModeRanges  = %s'    , [ CheckModes  , ModeRangesAsString  ] ));
          Add( IndFmt( anInd, '%s  ModeTypes   = %s'    , [ CheckModes  , ModeTypesAsString   ] ));
          Add( IndFmt( anInd,   '] // Knowledge'        , [                                   ] ));
        end;
        Add( IndFmt( anInd, 'TypeId      = %d'    , [ TypeId      ] ));
        Add( IndFmt( anInd, 'LongName    = ''%s''', [ LongName    ] ));
        Add( IndFmt( anInd, 'Name        = ''%s''', [ Name        ] ));
        Add( IndFmt( anInd, 'ModuleIndex = %d'    , [ ModuleIndex ] ));
        Add( IndFmt( anInd, 'Col         = %d'    , [ Col         ] ));
        Add( IndFmt( anInd, 'Row         = %d'    , [ Row         ] ));
        Add( IndFmt( anInd, 'Color       = %d'    , [ Color       ] ));
        Add( IndFmt( anInd, 'Unknown1    = $%.2x' , [ Unknown1    ] ));
        Add( IndFmt( anInd, 'Unknown2    = $%.1x' , [ Unknown2    ] ));
        DumpModes( aStrings, anInd);
      end;
    end;

    function    TModule.ModuleInfo: string; // override;
    begin
      Result := Format( '%s (%s)', [ ShortName, Name]);
    end;



{ ========
  TModuleList = class( TPatchChunk)
  private
    FKnowledge   : TModuleKnowledgeTable;
    FLocation    : TBits2;
    FModuleCount : TBits8;
    FModules     : array of TModule;
  public
    property    Knowledge: TModuleKnowledgeTable read FKnowledge write FKnowledge;
  public
    property    ModuleCount               : TBits8   read FModuleCount write SetModuleCount;
    property    Modules[ anIndex: Integer]: TModule read GetModule    write SetModule;
  private
}

    procedure   TModuleList.SetModuleCount( aValue: TBits8);
    var
      i : Integer;
    begin
      if aValue <> FModuleCount
      then begin
        for i := aValue to FModuleCount - 1
        do Modules[ i] := nil;
        SetLength( FModules, aValue);
        FModuleCount := aValue;
      end;
    end;

    function    TModuleList.GetModule( anIndex: Integer): TModule;
    begin
      Result := FModules[ anIndex];
    end;

    procedure   TModuleList.SetModule( anIndex: Integer; const aValue: TModule);
    begin
      if Assigned( FModules[ anIndex])
      then FModules[ anIndex].Free;
      FModules[ anIndex] := aValue;
    end;

    procedure   TModuleList.DeleteModules;
    begin
      ModuleCount := 0;
    end;

//  protected

    function    TModuleList.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      i           : Integer;
      aModuleType : TBits8;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
          Location    := Read( 2);
          ModuleCount := Read( 8);
          for i := 0 to ModuleCount - 1
          do begin
            aModuleType := Bits[ ReadPointer, 8];
            Modules[ i] := CreateModule( Self, aModuleType);
            Modules[ i].Knowledge := Knowledge.ModuleKnowledge[ aModuleType];
            with Modules[ i]
            do begin
              Result := ReadBits( aBitHose);   //   and then parse those bits.
              if not Result
              then Abort;                      // Jump to exception handler
            end;
          end;
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TModuleList.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    destructor  TModuleList.Destroy; // override;
    begin
      DeleteModules;
      inherited;
    end;

    procedure   TModuleList.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    var
      i : Integer;
    begin
      inherited DumpDetails( aStrings, anInd, aShowBinary, aShowKnowledge);
      with aStrings
      do begin
        Add( IndFmt( anInd, 'Location    = %s', [ Locations( FLocation) ]));
        Add( IndFmt( anInd, 'ModuleCount = %d', [ ModuleCount           ]));
        Add( IndFmt( anInd, 'Modules['        , [                       ]));
        for i := 0 to ModuleCount - 1
        do Modules[ i].DumpToStrings( aStrings, anInd + 1, aShowBinary, aShowKnowledge);
        Add( IndFmt( anInd, '] // Modules'    , [                       ]));
      end;
    end;

    function    TModuleList.FindModule( aModuleIndex: Integer): TModule;
    var
      i : Integer;
    begin
      Result := nil;
      for i := 0 to ModuleCount - 1
      do
        if Modules[ i].ModuleIndex = aModuleIndex
        then begin
          Result := Modules[ i];
          Break;
        end;
    end;


{ ========
  TUnknown1 = class( TPatchChunk)
  protected
}

    function    TUnknown1.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TUnknown1.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TUnknown1.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TCable = class( TPatchData)
  private
    FColor         : TBits3;
    FModuleFrom    : TBits8;
    FConnectorFrom : TBits6;
    FLinkType      : TBits1;
    FModuleTo      : TBits8;
    FConnectorTo   : TBits6;
  public
    property    Color         : TBits3 read FColor         write FColor        ;
    property    ModuleFrom    : TBits8 read FModuleFrom    write FModuleFrom   ;
    property    ConnectorFrom : TBits6 read FConnectorFrom write FConnectorFrom;
    property    LinkType      : TBits1 read FLinkType      write FLinkType     ;
    property    ModuleTo      : TBits8 read FModuleTo      write FModuleTo     ;
    property    ConnectorTo   : TBits6 read FConnectorTo   write FConnectorTo  ;
  protected
}

    function    TCable.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := False;
      try
        with aBitHose
        do begin
          Color         := Read( 3);
          ModuleFrom    := Read( 8);
          ConnectorFrom := Read( 6);
          LinkType      := Read( 1);
          ModuleTo      := Read( 8);
          ConnectorTo   := Read( 6);
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TCable.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    function    TCable.FindModule( aModuleIndex: TBits8): TModule;
    var
      apd   : TPatchData;
      asect : TBits2;
    begin
      apd   := Parent;
      asect := ( apd as TCableList).Location;
      while Assigned( apd) and not ( apd is TPatch)
      do apd := apd.Parent;
      if Assigned( apd)
      then begin
        case asect of
          0  : Result := TPatch( apd).FModuleListFX.FindModule( aModuleIndex);
          1  : Result := TPatch( apd).FModuleListVA.FindModule( aModuleIndex);
          else Result := nil;
        end;
      end
      else Result := nil;
    end;

    function    TCable.FindModuleInfo( aModuleIndex: TBits8): string;
    var
      aModule : TModule;
    begin
      aModule := FindModule( aModuleIndex);
      if Assigned( aModule)
      then Result := aModule.ModuleInfo
      else Result := '<no_info>';
    end;

    function    TCable.FindConnectorInfo(
      aModuleIndex   : TBits8;
      aConnectorIndex: TBits6;
      aLinkType      : TBits1;
      isFrom         : Boolean
    ): string;
    var
      aModule : TModule;
    begin
      aModule := FindModule( aModuleIndex);
      if Assigned( aModule)
      then Result := aModule.ConnectorInfo( aConnectorIndex, ( aLinkType = 0) or not isFrom)
      else Result := '<no_info>';
    end;

    procedure   TCable.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      with aStrings
      do begin
        Add(
          IndFmt(
            anInd,
            'CableType     = ''%s'''      ,
            [
              CableTypes ( LinkType)
            ]
          )
        );
        Add(
          IndFmt(
            anInd,
            'Color         = ''%s'''      ,
            [
              CableColors( Color   )
            ]
          )
        );
        Add(
          IndFmt(
            anInd,
            'ModuleFrom    = %3d    // %s',
            [
              ModuleFrom,
              FindModuleInfo( Modulefrom)
            ]
          )
        );
        Add(
          IndFmt(
            anInd,
            'ConnectorFrom = %3d    // %s',
            [
              ConnectorFrom,
              FindConnectorInfo( ModuleFrom, ConnectorFrom, LinkType, True )
            ]
          )
        );
        Add(
          IndFmt(
            anInd,
            'ModuleTo      = %3d    // %s',
            [
              ModuleTo,
              FindModuleInfo( ModuleTo)
            ]
          )
        );
        Add(
          IndFmt(
            anInd,
            'ConnectorTo   = %3d    // %s',
            [
              ConnectorTo,
              FindConnectorInfo( ModuleTo, ConnectorTo, LinkType, False)
            ]
          )
        );
      end;
    end;


{ ========
  TCableList = class( TPatchChunk)
  private
    FLocation   : TBits2;
    FCableCount : TBits8;
    FCables     : array of TCable;
  public
    property    Location                 : TBits2  read FLocation   write FLocation;
    property    CableCount               : TBits10 read FCableCount write SetCableCount;
    property    Cables[ anIndex: Integer]: TCable  read GetCable    write SetCable;
  private
}

    procedure   TCableList.SetCableCount( aValue: TBits10);
    var
      i : Integer;
    begin
      if aValue <> FCableCount
      then begin
        for i := aValue to FCableCount - 1
        do Cables[ i] := nil;
        SetLength( FCables, aValue);
        FCableCount := aValue;
      end;
    end;

    function    TCableList.GetCable( anIndex: Integer): TCable;
    begin
      Result := FCables[ anIndex];
    end;

    procedure   TCableList.SetCable( anIndex: Integer; const aValue: TCable);
    begin
      if Assigned( FCables[ anIndex])
      then FCables[ anIndex].Free;
      FCables[ anIndex] := aValue;
    end;

    procedure   TCableList.DeleteCables;
    begin
      CableCount := 0;
    end;

//  protected

    function    TCableList.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      i : Integer;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
          Location   := Read(  2);
          CableCount := Read( 22);
          for i := 0 to CableCount - 1
          do begin
            Cables[ i] := TCable.Create( Self);
            with Cables[ i]
            do begin
              Result := ReadBits( aBitHose); // Parse out a cable
              if not Result
              then Abort;                    // Jump to exception handler
            end;
          end;
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TCableList.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    destructor  TCableList.Destroy; // override;
    begin
      DeleteCables;
      inherited;
    end;

    procedure   TCableList.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    var
      i : Integer;
    begin
      inherited DumpDetails( aStrings, anInd, aShowBinary, aShowKnowledge);
      with aStrings
      do begin
        Add( IndFmt( anInd, 'Location   = %s' , [ Locations( FLocation)       ]));
        Add( IndFmt( anInd, 'Unknown    = #%s', [ LongToBinary( FUnknown, 12) ]));
        Add( IndFmt( anInd, 'CableCount = %d' , [ CableCount                  ]));
        Add( IndFmt( anInd, 'Cables['         , [                             ]));
        for i := 0 to CableCount - 1
        do Cables[ i].DumpToStrings( aStrings, anInd + 1, aShowBinary, aShowKnowledge);
        Add( IndFmt( anInd, '] // Cables'     , [                             ]));
      end;
    end;


{ ========
  TPatchSettings = class( TPatchChunk)
  protected
}

    function    TPatchSettings.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TPatchSettings.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TPatchSettings.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TParams = class( TPatchData)
  // Set of parameter values for one module one variation
  private
    FVariation   : Byte;
    FParamValues : array of TBits7;
  public
    property    Variation  : Byte    read FVariation    write FVariation;
    property    ParamCount : Integer read GetParamCount write SetParamCount;
    property    ParamValue[ anIndex: Integer]: TBits7 read GetParamValue write SetParamValue;
    property    ParamName [ anIndex: Integer]: string read GetParamName;
  private
}

    function    TParams.GetParamCount: Integer;
    begin
      Result := Length( FParamValues);
    end;

    procedure   TParams.SetParamCount( aValue: Integer);
    begin
      SetLength( FParamValues, aValue);
    end;

    function    TParams.GetParamValue( anIndex: Integer): TBits7;
    begin
      if ( anIndex >= 0) and ( anIndex < ParamCount)
      then Result := FParamValues[ anIndex]
      else Result := 0;
    end;

    procedure   TParams.SetParamValue( anIndex: Integer; aValue: TBits7);
    begin
      if ( anIndex >= 0) and ( anIndex < ParamCount)
      then FParamValues[ anIndex] := aValue;
    end;

    function    TParams.GetParamName( anIndex: Integer): string;
    var
      aPatch       : TPatch;
      aLocation    : Integer;
      aModuleIndex : Integer;
      aParent      : TPatchData;
      aModule      : TModule;
    begin
      Result  := Format( 'param_not_found (%d)', [ anIndex]);
      aParent := Parent;
      if aParent is TParamSet
      then begin
        aModuleIndex := TParamSet( aParent).ModuleIndex;
        aParent := aParent.Parent;
        if aParent is TModuleParameters
        then begin
          aLocation := TModuleParameters( aParent).Location;
          aParent := aParent.Parent;
          if aParent is TPatch
          then begin
            aPatch  := TPatch( aParent);
            aModule := aPatch.SearchModule( aLocation, aModuleIndex);
            if Assigned( aModule)
            then result := aModule.ParamName[ anIndex];
          end;
        end;
      end;
    end;

    function    TParams.GetParamRange( anIndex: Integer): string;
    var
      aPatch       : TPatch;
      aLocation    : Integer;
      aModuleIndex : Integer;
      aParent      : TPatchData;
      aModule      : TModule;
    begin
      Result  := Format( 'param_not_found (%d)', [ anIndex]);
      aParent := Parent;
      if aParent is TParamSet
      then begin
        aModuleIndex := TParamSet( aParent).ModuleIndex;
        aParent := aParent.Parent;
        if aParent is TModuleParameters
        then begin
          aLocation := TModuleParameters( aParent).Location;
          aParent := aParent.Parent;
          if aParent is TPatch
          then begin
            aPatch  := TPatch( aParent);
            aModule := aPatch.SearchModule( aLocation, aModuleIndex);
            if Assigned( aModule)
            then result := aModule.Knowledge.ParamRange[ anIndex];
          end;
        end;
      end;
    end;

//  protected

    function    TParams.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      i : Integer;
    begin
      Result := False;
      try
        with aBitHose
        do begin
          Variation := Read( 8);
          for i := 0 to ParamCount - 1
          do ParamValue[ i] := Read( 7);
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TParams.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TParams.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    var
      i : Integer;
    begin
      with aStrings
      do begin
        Add( IndFmt(    anInd, 'Variation = %s'                , [ VariationName( Variation and RA_MaskBits4) ]));
        Add( IndFmt(    anInd, 'ParamValues['                  , [                                            ]));
        for i := 0 to ParamCount - 1
        do
          Add(
            IndFmt(
              anInd,
              '  ParmVal %2d = %3d // %s :: %s',
              [
                i,
                ParamValue[ i],
                ExtendStr( ParamName[ i], 12),
                ParamRange[ i]
              ]
            )
          );
        Add( IndFmt(    anInd, '] // ParamValues', [                                                          ]));
      end;
    end;



{ ========
  TParamsSet = class( TPatchData)
  // Set of parameter values for one module all variations
  private
    // 0 .. 7 : variations 1 .. 8, 8 : init variation
    FVariations     : array[ 0 .. 8] of TParams;
    FModuleIndex    : TBits8;
    FParameterCount : Integer;
  public
    property    ModuleIndex    : TBits8  read FModuleIndex    write FModuleIndex;
    property    ParameterCount : Integer read FParameterCount write FParameterCount;
    property    VariationCount : Integer read GetVariationCount;
    property    Variation[ anIndex: Integer]: TParams read GetVariation write SetVariation;
  private
}

    function    TParamSet.GetVariationCount: Integer;
    begin
      Result := Length( FVariations);
    end;

    function    TParamSet.GetVariation( anIndex: Integer): TParams;
    begin
      if ( anIndex >= Low( FVariations)) and ( anIndex <= High( FVariations))
      then Result := FVariations[ anIndex]
      else Result := nil;
    end;

    procedure   TParamSet.SetVariation( anIndex: Integer; aValue: TParams);
    begin
      if ( anIndex >= Low( FVariations)) and ( anIndex <= High( FVariations))
      then begin
        if Assigned( FVariations[ anIndex])
        then FVariations[ anIndex].Free;
        FVariations[ anIndex] := aValue;
      end;
    end;

  // protected

    function    TParamSet.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      i : Integer;
    begin
      Result := False;
      try
        with aBitHose
        do begin
          ModuleIndex    := Read( 8);
          ParameterCount := Read( 7);
          for i := Low( FVariations) to High( FVariations)
          do begin
            FVariations[ i] := TParams.Create( Self);
            with FVariations[ i]
            do begin
              ParamCount := Self.ParameterCount;
              Result := ReadBits( aBitHose);   // and let ParamSet parse itself
              if not Result
              then Abort;                      // Jump to exception handler
            end;
          end;
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TParamSet.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    destructor  TParamSet.Destroy; // override;
    var
      i : Integer;
    begin
      for i := Low( FVariations) to High( FVariations)
      do FVariations[ i] := nil; // Freeing existing variations
      inherited;
    end;

    procedure   TParamSet.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    var
      i : Integer;
    begin
      // inherited DumpDetails( aStrings, anInd, aShowBinary, aShowKnowledge);
      with aStrings
      do begin
        Add( IndFmt( anInd, 'ModuleIndex    = %d // %s' , [ ModuleIndex, FindModuleInfo]));
        Add( IndFmt( anInd, 'ParameterCount = %d'       , [ ParameterCount ]));
        Add( IndFmt( anInd, 'Variations['    , [ Indent( anInd) ]));
        for i := Low( FVariations) to High( FVariations)
        do
          if Assigned( Variation[ i])
          then Variation[ i].DumpToStrings( aStrings, anInd + 1, aShowBinary, aShowKnowledge);
        Add( IndFmt( anInd, '] // Variations', []));
      end;
    end;

    function    TParamSet.FindModule: TModule;
    var
      aPD   : TPatchData;
      aSect : TBits2;
    begin
      aPD   := Parent;
      aSect := ( aPD as TModuleParameters).Location;

      while Assigned( aPD) and not ( aPD is TPatch)
      do aPD := aPD.Parent;

      if aPD is TPatch
      then
        with TPatch( aPD)
        do
          case aSect of
            0  : Result := FModuleListFX.FindModule( ModuleIndex);
            1  : Result := FModuleListVA.FindModule( ModuleIndex);
            else Result := nil;
          end
      else Result := nil;
    end;


    function    TParamSet.FindModuleInfo: string;
    var
      aModule: TModule;
    begin
      aModule := FindModule;
      if Assigned( aModule)
      then Result := aModule.ModuleInfo
      else result := 'no_info_found';
    end;


{ ========
  TModuleParameters = class( TPatchChunk)
  private
    FLocation  : Tbits2;
    FSetCount  : TBits8;
    FUnknown   : TBits8;
    FParamSets : array of TParamSet;
  public
    property    Location : TBits2 read FLocation   write FLocation;
    property    SetCount : TBits8 read GetSetCount write SetSetCount;
    property    Unknown  : TBits8 read FUnknown    write FUnknown;
    property    ParamSet[ anIndex: TBits8]: TParamSet read GetParamSet write SetParamSet;
  private
}

    function    TModuleParameters.GetSetCount: TBits8;
    begin
      Result := Length( FParamSets);
    end;

    procedure   TModuleParameters.SetSetCount( aValue: TBits8);
    var
      i         : Integer;
      aParamSet : TParamSet;
    begin
      for i := aValue to SetCount - 1
      do
        if Assigned( FParamSets[ i])
        then begin
          aParamSet := FParamSets[ i];
          FParamSets[ i] := nil;
          aParamSet.Free;
        end;
      SetLength( FParamSets, aValue);
    end;

    function    TModuleParameters.GetParamSet( anIndex: TBits8): TParamSet;
    begin
      if {( anIndex >= 0) and} ( anIndex < SetCount)
      then Result := FParamSets[ anIndex]
      else Result := nil;
    end;

    procedure   TModuleParameters.SetParamSet( anIndex: TBits8; aValue: TParamSet);
    begin
      if {( anIndex >= 0) and} ( anIndex < SetCount)
      then FParamSets[ anIndex] := aValue;
    end;

//  protected

    function    TModuleParameters.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
     i : Integer;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
          Location := Read( 2);
          SetCount := Read( 8);
          UnKnown  := Read( 8);
          for i := 0 to SetCount - 1
          do begin
            ParamSet[ i] := TParamSet.Create( Self);
            ParamSet[ i].ModuleIndex    := Bits[ ReadPointer + 0, 8];
            ParamSet[ i].ParameterCount := Bits[ ReadPointer + 8, 7];
            with ParamSet[ i]
            do begin
              Result := ReadBits( aBitHose);            // and let it parse itself
              if not Result
              then Abort;                               // Jump to exception handler
            end;
          end;
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TModuleParameters.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    destructor  TModuleParameters.Destroy; // override;
    begin
      SetCount := 0;
      inherited;
    end;

    procedure   TModuleParameters.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    var
      i : Integer;
    begin
      inherited DumpDetails( aStrings, anInd, aShowBinary, aShowKnowledge);
      with aStrings
      do begin
        Add( IndFmt( anInd, 'Location = %s' , [ Locations( FLocation)     ]));
        Add( IndFmt( anInd, 'SetCount = %d' , [ SetCount                  ]));
        Add( IndFmt( anInd, 'Unknown  = #%s', [ LongToBinary( Unknown, 8) ]));
        Add( IndFmt( anInd, 'ParamSets['    , [                           ]));
        for i := 0 to SetCount - 1
        do ParamSet[ i].DumpToStrings( aStrings, anInd + 1, aShowBinary, aShowKnowledge);
        Add( IndFmt( anInd, '] // ParamSets', [                           ]));
      end;
    end;


{ ========
  TUnknown2 = class( TPatchChunk)
  protected
}

    function    TUnknown2.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TUnknown2.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TUnknown2.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TKnobAssignments = class( TPatchChunk)
  protected
}

    function    TKnobAssignments.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TKnobAssignments.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TKnobAssignments.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TControllerAssignments = class( TPAtchChunk)
  protected
}

    function    TControllerAssignments.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TControllerAssignments.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TControllerAssignments.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TUnknown3 = class( TPatchChunk)
  protected
}

    function    TUnknown3.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TUnknown3.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TUnknown3.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TName = class( TPatchData)
  private
    FModuleIndex : TBits8;
    FName        : string;
  public
    property    ModuleIndex: TBits8 read FModuleIndex write FModuleIndex;
    property    Name       : string read FName        write FName;
  protected
}

    function    TName.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := False;
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TName.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;


{ ========
  TModuleNames = class( TPatchChunk)
  private
    FLocation  : TBits2;
    FUnknown   : TBits6;
    FNameCount : TBits8;
    FNames     : array of TName;
  public
    property    Location  : TBits2 read FLocation    write FLocation;
    property    Unknown   : Tits6  read FUnknown     write FUnknown;
    property    NameCount: Integer read GetNameCount write SetNameCount;
    property    Name[ aModuleIndex: TBits8]: string read GetName write SetName;
  private
}

    function    TModuleNames.GetNameCount: Integer;
    begin
      Result := Length( FNames);
    end;

    procedure   TModuleNames.SetNameCount( aValue: Integer);
    var
      i     : Integer;
      aName : TName;
    begin
      for i := aValue to NameCount - 1
      do
        if Assigned( FNames[ i])
        then begin
          aName := FNames[ i];
          FNames[ i] := nil;
          aName.Free;
        end;
      SetLength( FNames, aValue);
    end;

    function    TModuleNames.GetName( aModuleIndex: TBits8): string;
    var
      i : Integer;
    begin
      Result := 'invalid_name';
      for i := 0 to NameCount - 1
      do
        if FNames[ i].ModuleIndex = aModuleIndex
        then begin
          Result := FNames[ i].Name;
          Break;
        end;
    end;

    procedure   TModuleNames.SetName( aModuleIndex: TBits8; const aValue: string);
    var
      i : Integer;
    begin
      for i := 0 to NameCount - 1
      do
        if FNames[ i].ModuleIndex = aModuleIndex
        then begin
          FNames[ i].Name := aValue;
          Break;
        end;
    end;

//  private

    procedure   TModuleNames.AddName( aModuleIndex: TBits8; const aName: string);
    begin
      NameCount := NameCount + 1;
      FNames[ NameCount - 1] := TName.Create( Self);
      with FNames[ NameCount - 1]
      do begin
        ModuleIndex := aModuleIndex;
        Name        := aName;
      end;
    end;

//  protected

    function    TModuleNames.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    const
      MaxLen = 16;
    var
      i            : Integer;
      aChar        : TBits8;
      aName        : string;
      aModuleIndex : TBits8;
      aNameCount   : TBits8;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
          Location   := Read( 2);
          Unknown    := Read( 6);
          aNameCount := Read( 8);
          for i := 0 to aNameCount - 1
          do begin
            aModuleIndex := Read( 8);
            aName        := '';
            aChar := Read( 8);
            while aChar <> 0
            do begin
              aName := aName + Char( aChar);
              if Length( aName) = MaxLen
              then Break
              else aChar := Read( 8);
            end;
            AddName( aModuleIndex, aName);
          end;
          Result := NameCount = aNameCount;
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TModuleNames.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    destructor  TModuleNames.Destroy; // override;
    begin
      NameCount := 0;
      inherited;
    end;

    procedure   TModuleNames.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    var
      i : Integer;
    begin
      inherited DumpDetails( aStrings, anInd, aShowBinary, aShowKnowledge);
      with aStrings
      do begin
        Add( IndFmt( anInd,   'Location  = %s'        , [ Locations( FLocation)     ]));
        Add( IndFmt( anInd,   'Unknown   = #%s'       , [ LongToBinary( Unknown, 6) ]));
        Add( IndFmt( anInd,   'NameCount = %d'        , [ NameCount                 ]));
        Add( IndFmt( anInd,   'Names['                , [                           ]));
        for i := 0 to NameCount - 1
        do begin
          Add( IndFmt( anInd, '  ModuleIndex = %d'    , [ FNames[ i].ModuleIndex    ]));
          Add( IndFmt( anInd, '  Name        = ''%s''', [ FNames[ i].Name           ]));
        end;
        Add( IndFmt( anInd,   '] // Names'            , [                           ]));
      end;
    end;


{ ========
  TTextPad = class( TPatchChunk)
  protected
}

    function    TTextPad.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := inherited ReadBits( aBitHose);
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TTextPad.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    procedure   TTextPad.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
    end;


{ ========
  TAbstractPatch = class( TPatchData)
  private
    FStrings   : TStringList;
    FPatchName : string;
    FVersion   : Byte;
    FPatchType : Byte;
  public
    property    PatchName: string read FpatchName write FPatchName;
    property    Version  : Byte   read FVersion   write FVersion;
    property    PatchType: Byte   read FPatchType write FPatchType;
  protected
}


    function    TAbstractPatch.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := False;
      try
        with aBitHose
        do begin
          Version   := Read( 8);
          PatchType := Read( 8);
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TAbstractPatch.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

//  public

    constructor TAbstractPatch.Create( aParent: TPatchData);
    begin
      inherited;
      FStrings := TStringList.Create;
      FBits    := TBitHose.Create;
    end;

    destructor  TAbstractPatch.Destroy; // override;
    begin
      FreeAndNil( FBits);
      FreeAndNil( FStrings);
      inherited;
    end;

//  public

    function    TAbstractPatch.LoadFromStream( const aStream: TStream): Boolean;
    var
      aCrc  : Word;
      aChar : Char;
      B     : Byte;
      S     : string;
    begin
      Result := False;
      if aStream.Size < 100 * 1024     // Skip files over 100 k
      then begin
        S := '';
        FStrings.Clear;
        with aStream do
        begin
          Position := 0;
          while Position < Size
          do begin
            Read( aChar, SizeOf( aChar));
            case aChar of
              #$0d : { just skip it } ;
              #$0a : begin
                  FStrings.Add( S);
                  S := '';
                end;
              #$00 : Break;
              else S := S + aChar;
            end;
          end;
          if S <> ''
          then FStrings.Add( S);
          aCrc := 0;
          FBits.Clear;
          while Position < Size
          do begin
            Read( B, SizeOf( B));
            aCrc := CrcClavia( aCrc, B);
            FBits.Write( 8, B);
          end;
        end;
        Result := aCrc = 0;
      end;
      if Result
      then begin
        FBits.Reset;
        Result := ReadBits( FBits);
      end;
    end;

    function    TAbstractPatch.SavetoStream( const aStream: TStream): Boolean;
    var
      S     : string;
      i     : Integer;
      aChar : Char;
      B     : Byte;
      aCrc  : Word;
    begin
      Result := False;
      FBits.Clear;
      WriteBits( FBits);
      try
        with aStream do
        begin
          Position := 0;
          with FStrings do
          begin
            for i := 0 to Count - 1 do
            begin
              S := Strings[ i];
              Write( S[ 1], Length( S));
              aChar := #$0d;
              Write( aChar, SizeOf( aChar));
              aChar := #$0a;
              Write( aChar, SizeOf( aChar));
            end;
          end;
          aChar := #0;
          Write( aChar, SizeOf( aChar));
          aCrc := 0;
          with FBits do
          begin
            while ReadPointer < Size
            do begin
              B    := Read( 8);
              aCrc := CrcClavia( aCrc, B);
              Write( B, SizeOf( B));
            end;
          end;
          B := ( aCrc shr 8) and $ff;
          Write( B, SizeOf( B));
          B := aCrc and $ff;
          Write( B, SizeOf( B));
          Result := True;
        end;
      except
        // ignore exceptions, just return False
      end;
    end;

    function    TAbstractPatch.LoadFromFile( const aFileName : string): Boolean;
    var
      aStream : TFileStream;
    begin
      Result := False;
      try
        if FileExists( aFileName)
        then begin
          PatchName := aFileName;
          aStream   := TFileStream.Create( aFileName, fmOpenRead + fmShareDenyWrite);
          try
            Result := LoadFromStream( aStream);
          finally
            aStream.Free;
          end;
        end;
      except
        // ignore exceptions, just return False
      end;
    end;

    function    TAbstractPatch.SavetoFile( const aFileName : string): Boolean;
    var
      aStream : TFileStream;
    begin
      Result := False;
      try
        PatchName := aFileName;
        aStream   := TFileStream.Create( aFileName, fmCreate);
        try
          Result := SaveToStream( aStream);
        finally
          aStream.Free;
        end;
      except
        // ignore exceptions, just return False
      end;
    end;

//  public

    procedure   TAbstractPatch.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      with aStrings
      do begin
        Add( IndFmt( anInd, 'PatchName = %s'  , [ PatchName              ]));
        Add( IndFmt( anInd, 'Version   = %d'  , [ Version                ]));
        Add( IndFmt( anInd, 'PatchType = %s'  , [ PatchTypes( PatchType) ]));
      end;
    end;


{ ========
  TPatch = class( TPatchData)
  private
    FPatchDescription      : TPatchDescription;
    FModuleListVA          : TModuleList;
    FModuleListFX          : TModuleList;
    FUnknown1              : TUnknown1;
    FCableListVA           : TCableList;
    FCableListFX           : TCableList;
    FPatchSettings         : TPatchSettings;
    FModuleParametersVA    : TModuleParameters;
    FModuleParametersFX    : TModuleParameters;
    FUnknown2              : TUnknown2;
    FKnobAssignments       : TKnobAssignments;
    FControllerAssignments : TControllerAssignments;
    FUnknown3_1            : TUnknown3;
    FUnknown3_2            : TUnknown3;
    FUnknown3_3            : TUnknown3;
    FModuleNamesVA         : TModuleNames;
    FModuleNamesFX         : TModuleNames;
    FTextPad               : TTextPad;
  protected
}

    function    TPatch.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    var
      anId    : Byte;
      aStart  : Integer;
      aLength : Integer;
      aChunk  : TPatchChunk;
    begin
      Result := False;
      FreeChunks;
      try
        Result := inherited ReadBits( aBitHose);
        if Result
        then begin
          with aBitHose
          do begin
            while Result and ( ReadPointer < Size - 16) // Don't scan over CRC.
            do begin
              aChunk  := nil;
              anId    := Bits[ ReadPointer + 0,  8];
              aLength := Bits[ ReadPointer + 8, 16];
              aStart  := ReadPointer + 24;
              case anId of
                $21 : begin
                    FPatchDescription := TPatchDescription.Create( Self);
                    aChunk            := FPatchDescription;
                  end;
                $4a : begin // multiple
                    case Bits[ ReadPointer + 24, 2] of
                      0 : begin
                          FModuleListFX := TModuleList.Create( Self);
                          aChunk        := FModuleListFX;
                          FModuleListFX.Knowledge := ModuleKnowledgeTable;
                        end;
                      1 : begin
                          FModuleListVA := TModuleList.Create( Self);
                          aChunk        := FModuleListVA;
                          FModuleListVA.Knowledge := ModuleKnowledgeTable;
                        end;
                      else Result := False; // Unhandled section, break out
                    end;
                  end;
                $69 : begin
                    FUnknown1 := TUnknown1.Create( Self);
                    aChunk    := FUnknown1;
                  end;
                $52 : begin // multiple
                    case Bits[ ReadPointer + 24, 2] of
                      0 : begin
                          FCableListFX := TCableList.Create( Self);
                          aChunk       := FCableListFX;
                        end;
                      1 : begin
                          FCableListVA := TCableList.Create( Self);
                          aChunk       := FCableListVA;
                        end;
                      else Result := False; // Unhandled section, break out
                    end;
                  end;
                $4d : begin // multiple
                    case Bits[ ReadPointer + 24, 2] of
                      0 : begin
                          FModuleParametersFX := TModuleParameters.Create( Self);
                          aChunk              := FModuleParametersFX;
                        end;
                      1 : begin
                          FModuleParametersVA := TModuleParameters.Create( Self);
                          aChunk              := FModuleParametersVA;
                        end;
                      2 : begin
                          FPatchSettings := TPatchSettings.Create( Self);
                          aChunk         := FPatchSettings;
                        end;
                      else Result := False; // Unhandled section, break out
                    end;
                  end;
                $65 : begin
                    FUnknown2 := TUnknown2.Create( Self);
                    aChunk    := FUnknown2;
                  end;
                $62 : begin
                    FKnobAssignments := TKnobAssignments.Create( Self);
                    aChunk           := FKnobAssignments;
                  end;
                $60 : begin
                    FControllerAssignments := TControllerAssignments.Create( Self);
                    aChunk                 := FControllerAssignments;
                  end;
                $5b : begin // multiple
                    if not Assigned( FUnknown3_1)
                    then begin
                      FUnknown3_1 := TUnknown3.Create( Self);
                      aChunk      := FUnknown3_1;
                    end
                    else if not Assigned( FUnknown3_2)
                    then begin
                      FUnknown3_2 := TUnknown3.Create( Self);
                      aChunk      := FUnknown3_2;
                    end
                    else if not Assigned( FUnknown3_3)
                    then begin
                      FUnknown3_3 := TUnknown3.Create( Self);
                      aChunk      := FUnknown3_3;
                    end
                    else Result := False;   // Unhandled section, break out
                  end;
                $5a : begin // multiple
                    case Bits[ ReadPointer + 24, 2] of
                      0 : begin
                          FModuleNamesFX := TModuleNames.Create( Self);
                          aChunk         := FModuleNamesFX;
                        end;
                      1 : begin
                          FModuleNamesVA := TModuleNames.Create( Self);
                          aChunk         := FModuleNamesVA;
                        end;
                      else Result := False; // Unhandled section, break out
                    end;
                  end;
                $6f : begin
                    FTextPad := TTextPad.Create( Self);
                    aChunk   := FTextPad;
                  end;
                else Result := False;       // Unhandled section, break out.
              end;
              if not Assigned( aChunk)
              then Result := False;
              if Result
              then begin
                with aChunk
                do begin
                  TypeId     := anId;
                  DataLength := aLength;
                  Result     := ReadBits( aBitHose);
                end;
              end;
              ReadPointer := aStart + 8 * aLength; // Align read pointer for next chunk.
            end;
          end;
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TPatch.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

    procedure   TPatch.FreeChunks;
    begin
      if Assigned( FPatchDescription     ) then FreeAndNil( FPatchDescription);
      if Assigned( FModuleListVA         ) then FreeAndNil( FModuleListVA);
      if Assigned( FModuleListFX         ) then FreeAndNil( FModuleListFX);
      if Assigned( FUnknown1             ) then FreeAndNil( FUnknown1);
      if Assigned( FCableListVA          ) then FreeAndNil( FCableListVA);
      if Assigned( FCableListFX          ) then FreeAndNil( FCableListFX);
      if Assigned( FPatchSettings        ) then FreeAndNil( FPatchSettings);
      if Assigned( FModuleParametersVA   ) then FreeAndNil( FModuleParametersVA);
      if Assigned( FModuleParametersFX   ) then FreeAndNil( FModuleParametersFX);
      if Assigned( FUnknown2             ) then FreeAndNil( FUnknown2);
      if Assigned( FKnobAssignments      ) then FreeAndNil( FKnobAssignments);
      if Assigned( FControllerAssignments) then FreeAndNil( FControllerAssignments);
      if Assigned( FUnknown3_1           ) then FreeAndNil( FUnknown3_1);
      if Assigned( FUnknown3_2           ) then FreeAndNil( FUnknown3_2);
      if Assigned( FUnknown3_3           ) then FreeAndNil( FUnknown3_3);
      if Assigned( FModuleNamesVA        ) then FreeAndNil( FModuleNamesVA);
      if Assigned( FModuleNamesFX        ) then FreeAndNil( FModuleNamesFX);
      if Assigned( FTextPad              ) then FreeAndNil( FTextPad);
    end;

//  public

    constructor TPatch.Create( aParent: TPatchData);
    begin
      inherited;
    end;

    destructor  TPatch.Destroy; // override;
    begin
      FreeChunks;
      inherited;
    end;

    function    TPatch.SearchModule( aLocation, aModuleIndex: Integer): TModule;
    begin
      Result := nil;
      if ( aLocation = 0) and Assigned( FModuleListFX)
      then Result := FModuleListFX.FindModule( aModuleIndex) as TModule
      else if ( aLocation = 1) and Assigned( FModuleListVA)
      then Result := FModuleListVA.FindModule( aModuleIndex) as TModule;
    end;

    function    TPatch.SearchModuleName( aLocation, aModuleIndex: Integer): string;
    begin
      Result := 'module_name_not_found';
      if ( aLocation = 0) and Assigned( FModuleNamesFX)
      then Result := FModuleNamesFX.Name[ aModuleIndex]
      else if ( aLocation = 1) and Assigned( FModuleNamesVA)
      then Result := FModuleNamesVA.Name[ aModuleIndex];
    end;

    procedure   TPatch.SetModuleName( aLocation, aModuleIndex: Integer; const aValue: string);
    begin
      if ( aLocation = 0) and Assigned( FModuleNamesFX)
      then FModuleNamesFX.Name[ aModuleIndex] := aValue
      else if ( aLocation = 1) and Assigned( FModuleNamesVA)
      then FModuleNamesVA.Name[ aModuleIndex] := aValue;
    end;

//  public

    procedure   TPatch.DumpToStrings( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      aStrings.Clear;
      inherited;
    end;

    procedure   TPatch.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      inherited;
      if Assigned( FPatchDescription     ) then FPatchDescription     .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FModuleListVA         ) then FModuleListVA         .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FModuleListFX         ) then FModuleListFX         .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FUnknown1             ) then FUnknown1             .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FCableListVA          ) then FCableListVA          .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FCableListFX          ) then FCableListFX          .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FPatchSettings        ) then FPatchSettings        .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FModuleParametersVA   ) then FModuleParametersVA   .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FModuleParametersFX   ) then FModuleParametersFX   .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FUnknown2             ) then FUnknown2             .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FKnobAssignments      ) then FKnobAssignments      .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FControllerAssignments) then FControllerAssignments.DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FUnknown3_1           ) then FUnknown3_1           .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FUnknown3_2           ) then FUnknown3_2           .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FUnknown3_3           ) then FUnknown3_3           .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FModuleNamesVA        ) then FModuleNamesVA        .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FModuleNamesFX        ) then FModuleNamesFX        .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
      if Assigned( FTextPad              ) then FTextPad              .DumpToStrings( aStrings, anInd, aShowBinary, aShowKnowledge);
    end;


{ ========
  TPerformance = class( TAbstractPatch)
  private
    FPatchA : TPatch;
    FPatchB : TPAtch;
    FPatchC : TPatch;
    FPAtchD : TPatch;
  public
    property    PatchA : TPatch read FPatchA;
    property    PatchB : TPatch read FPatchB;
    property    PatchC : TPatch read FPatchC;
    property    PatchD : TPatch read FPatchD;
  protected
}

    function    TPerformance.ReadBits( const aBitHose: TBitHose): Boolean; //override;
    begin
      Result := False;
      try
        with aBitHose
        do begin
        end;
        Result := True;
      except
        // Kill all exceptions, just return false
      end;
    end;

    procedure   TPerformance.WriteBits( const aBitHose: TBitHose); // override;
    begin
    end;

    procedure   TPerformance.FreePatches;
    begin
      if Assigned( FPAtchA) then FreeAndNil( FPatchA);
      if Assigned( FPAtchB) then FreeAndNil( FPatchB);
      if Assigned( FPAtchC) then FreeAndNil( FPatchC);
      if Assigned( FPAtchD) then FreeAndNil( FPatchD);
    end;

//  public

    constructor TPerformance.Create( aParent: TPatchData);
    begin
      inherited;
    end;

    destructor  TPerformance.Destroy; // override;
    begin
      FreePatches;
      inherited;
    end;

//  public

    procedure   TPerformance.DumpToStrings( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); // override;
    begin
      aStrings.Clear;
      inherited;
    end;

    procedure   TPerformance.DumpDetails( const aStrings: TStrings; anInd: Integer; aShowBinary, aShowKnowledge: Boolean); //override;
    begin
      inherited;
    end;


end.

